#include <signal.h>
#include <time.h>
#include <errno.h>
#include <linux/unistd.h>
#include "posix_time.h"
#include "clock_realtime.h"
#include "clock_linux_counter.h"

static struct clock_func *posix_clock[NOF_CLOCKS] = {
  &clock_realtime,
  &clock_linux_counter
};

#define CLOCKID_SHIFT 16
#define TIMERID_MASK  ((1 << CLOCKID_SHIFT) - 1)

#define call_func(func, id, args) \
 ((id >= 0) && (id < NOF_CLOCKS) && \
  posix_clock[id] && posix_clock[id]->func) ? \
	posix_clock[id]->func args : \
	(errno = EINVAL, -1)


int clock_gettime(clockid_t which_clock, struct timespec *ts)
{
  return call_func(clock_gettime, which_clock, (which_clock, ts));
}

int clock_settime(clockid_t which_clock, const struct timespec *setting)
{
  return call_func(clock_settime, which_clock, (which_clock, setting));
}

int clock_getres(clockid_t which_clock, struct timespec *resolution)
{
  return call_func(clock_getres, which_clock, (which_clock, resolution));
}

int timer_create(clockid_t which_clock, struct sigevent *timer_event_spec,
                 timer_t *created_timer_id)
{
  int res;

  res = call_func(timer_create, which_clock,
		   (which_clock, timer_event_spec, created_timer_id));
  if (res == 0) {
    *created_timer_id |= (which_clock << CLOCKID_SHIFT);
  }
  return res;
}

int timer_gettime(timer_t timer_id, struct itimerspec *setting)
{
  return call_func(timer_gettime, timer_id >> CLOCKID_SHIFT,
		   (timer_id & TIMERID_MASK, setting));
}

int timer_settime(timer_t timer_id,
                  int flags,
                  const struct itimerspec *new_setting,
                  struct itimerspec *old_setting)
{
  return call_func(timer_settime, timer_id >> CLOCKID_SHIFT,
		   (timer_id & TIMERID_MASK, flags, new_setting, old_setting));
}

int timer_getoverrun(timer_t timer_id)
{
  return call_func(timer_getoverrun, timer_id >> CLOCKID_SHIFT, 
		   (timer_id & TIMERID_MASK));
}

int timer_delete(timer_t timer_id)
{
  return call_func(timer_delete, timer_id >> CLOCKID_SHIFT, 
		   (timer_id & TIMERID_MASK));
}

int clock_install(clockid_t which_clock, struct clock_func *new_clock,
		  struct clock_func **old_clock)
{
  if ((which_clock < 0) || (which_clock >= NOF_CLOCKS)) {
    errno = EINVAL;
    return -1;
  }
  if (old_clock) *old_clock = posix_clock[which_clock];
  if (new_clock) posix_clock[which_clock] = new_clock;

  return 0;
}
